/************************************************************************/
/*  qrequest -	create request record for specified channels.		*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: qrequestb4.c,v 1.2 2000/10/26 13:39:56 doug Exp $ ";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
   
#include "version.h"

char *syntax[] = {
"%s version " VERSION,
"%s  -f date_time [ -t date_time | -s span_interval ]",
"	[ -H holdings_file | -D holdings_directory ]",
"	[ -R reject_station_list ] [ -I reject_channel_list ]",
"	[ -d n ] [ -h ]	station channel_1 [ ... channel_n ]",
"  where:",
"	-H holdings_file",
"		    specifies the name of the holdings file.",
"	-D holdings_directory",
"		    specifies the name of the holdings directory.",
"		    Daily holdings files for day yyyy.doy should be named:",
#ifdef SHORTYEAR
"			holding_directory/yyyy/holdings.yy.doy",
#else
"			holding_directory/yyyy/holdings.yyyy.doy",
#endif
"	-R reject_station_list",
"		    specifies a comma-delimited list of stations to reject.",
"	-I reject_channel_list",
"		    specifies a comma-delimited list of channels to reject.",
"	-f date	    From date - ignore data before this date.",
"	-t date	    To date - ignore data from this date on.",
"		    Date can be in formats:",
"			[19]yy/mm/dd/hh:mm:ss.ffff",
"			[19]yy/mm/dd.hh:mm:ss.ffff",
"			[19]yy/mm/dd,hh:mm:ss.ffff",
"			[19]yy.ddd.hh:mm:ss.ffff",
"			[19]yy.ddd,hh:mm:ss.ffff",
"			[19]yy,ddd,hh:mm:ss.ffff",
"	-s interval span interval.  Alternate way of specifying end time.",
"		    Interval can be an integer followed immediately by",
"		    s, m, h, or d for seconds, minutes, hours, or days.",
"	-d n	    Debug output (OR of any of the following values:)",
"			1 = time slews",
"			2 = time info",
"			4 = block info",
"			8 = block headers",
"			16 = flags",
"			32 = input stream",
"			64 = output stream",
"	-h	    Help - prints syntax message.",
"	station	    specifies the station name.  It may contain quoted wildcard",
"		    characters and/or a comma-delimited list of stations.",
"	channel_i   specifies the channel(s).  At least 1 channel must be",
"		    specified. They may contain quoted wildcard characters.",
NULL };

/************************************************************************/

#define	EXTERN

#define	DEBUG
#define	DEBUG_SLEW	1
#define	DEBUG_TIME	2
#define	DEBUG_BLOCK	4
#define	DEBUG_HDRS	8
#define	DEBUG_FLAG	16
#define	DEBUG_STREAM	32
#define	DEBUG_OUT	64
#define	DEBUG_ANY	\
    (DEBUG_BLOCK | DEBUG_HDRS | DEBUG_SLEW | DEBUG_TIME | DEBUG_FLAG | \
     DEBUG_STREAM | DEBUG_OUT)
#define	debug(val)  (debug_option & (val))

#include "qlib2.h"

#include "holdings.h"
#include "procs.h"
#include "externals.h"

/*	Assumptions:	s1<=e1, s2<=e2					*/
#define	WINDOW_INTERSECT(s1,e1,s2,e2)    (!(e1<=s2 || s1>=e2))

#define	error(str)	{fprintf (stderr, "error: %s\n",str);}

char * gen_req (HOLDING_REC hr, INT_TIME req_start_time, INT_TIME req_end_time);

/************************************************************************/
/*  main:   main program.						*/
/************************************************************************/
main (int argc, char **argv)
{
    time_t	u_start, u_end;
    EXT_TIME	start_ext, end_ext;
    char	daily_file[1024];

    /*	Initialization.							*/
    info = stdout;
    date_fmt = JULIAN_FMT;
    debug_option = 0;
    get_my_wordorder();

    parse_cmdline (&argc, &argv);
    u_start = unix_time_from_int_time (start_time);
    u_end = unix_time_from_int_time (end_time);
    if (u_end <= u_start) {
	fprintf (stderr, "end time before start time\n");
	exit(1);
    }

    /*	Scan a single holdings file is holdings file is specified.	*/
    if (holdings_file) {
	/*	Open the holdings file.					*/
	if ( (hfp =fopen (holdings_file, "r")) == NULL) {
	    fprintf (stderr, "unable to open holdings file %s\n", holdings_file);
	    exit(1);
	}
	if (debug(DEBUG_STREAM)) fprintf (info, "input file: %s\n", holdings_file);
	scan_holdings(u_start, u_end);
    }

    /*	Scan daily files if holdings directory is specified.		*/
    if (holdings_dir) {
	end_ext = int_to_ext(end_time);
	while (tdiff (end_time, start_time) > 0.) {
	    start_ext = int_to_ext(start_time);
#ifdef SHORTYEAR
	    sprintf (daily_file, "%s/%04d/holdings.%02d.%03d", holdings_dir,
		     start_ext.year, start_ext.year%100, start_ext.doy);
#else
	    sprintf (daily_file, "%s/%04d/holdings.%04d.%03d", holdings_dir,
		     start_ext.year, start_ext.year, start_ext.doy);
#endif
	    if ( (hfp =fopen (daily_file, "r")) != NULL) {
		if (debug(DEBUG_STREAM)) fprintf (info, "input file: %s\n", daily_file);
		scan_holdings(u_start, u_end);
		fclose (hfp);
	    }
	    start_ext = next_day (start_ext);
	    start_time = ext_to_int(start_ext);
	}
    }
    return (0);
}

/************************************************************************/
/*  next_day:								*/
/*	Compute the beginning of the next day.				*/
/************************************************************************/
EXT_TIME next_day
   (EXT_TIME	et)		/* date/time in EXT_TIME format.	*/
{    
    et.hour = et.minute = et.second = et.usec = 0;
    et.doy++;
    if (et.doy > DAYS_PER_YEAR(et.year)) {
	et.year++;
	et.doy = et.month = et.day = 1;
    }
    else dy_to_mdy (et.doy, et.year, &(et.month), &(et.day));
    return (et);
}

/************************************************************************/
/*  scan_holdings:							*/
/*	Scan holdings and generate request(s) for the specified time.	*/
/************************************************************************/
int scan_holdings
   (time_t	u_start,	/* start of request time.		*/
    time_t	u_end)		/* end of request time.			*/
{
    
    int		i, j, k;
    int		matched;
    int		rejected;
    INT_TIME	req_start_time;
    INT_TIME	req_end_time;
    INT_TIME	hr_start_time, hr_end_time;
    time_t	hr_u_start, hr_u_end;
    SDR_TIME	sdr_time;

    char	line[256];
    char	hbuf[sizeof(HOLDING_REC) + 1];
    HOLDING_REC	hr;
    char	str[80], *p;
    char	hr_station[HR_STATION_LEN+1];
    char	hr_channel[HR_CHANNEL_LEN+1];
    INT_TIME	*pt;
    char	*hr_line;

    while (fgets((char *)hbuf,sizeof(hbuf),hfp) != NULL) {
	memcpy ((void *)&hr, hbuf, sizeof(hr));
	/*  Convert tab and newline delimiters to nulls to terminate the strings.*/
	for (i=0, p=(char *)&hr; i<sizeof(hr); i++, ++p) {
	    if (*p == '\t') *p = '\0';
	    if (*p == '\n') *p = '\0';
	}

	/* Look for a match on (station, channel, time_window).		*/
	matched = 0;
	trim(strcpy(hr_station, hr.station));
	for (i=0; i<n_stations; i++) {
	    if (matched || fnmatch(p_stations[i], hr_station, 0)!=0) continue;
	    /* We have a match on a station.			      	*/
	    /* Now check for a rejection match.				*/
	    rejected = 0;
	    for (k=0; k<n_srejects; k++) {
		if (fnmatch(p_srejects[k], hr_station, 0)==0) {
		    rejected++;
		}
	    }
	    if (rejected) continue;
	    for (j=0; j<n_channels; j++) {
		trim(strcpy(hr_channel, hr.channel));
		if (matched || fnmatch(p_channels[j], hr_channel, 0)!=0) continue;
		/* We have a channel match.				*/
		++matched;
		rejected = 0;
		/* Now check for a rejection match.			*/
		for (k=0; k<n_crejects; k++) {
		    if (fnmatch(p_crejects[k], hr_channel, 0)==0) {
			++rejected;
			break;
		    }
		}
		if (rejected) continue;
		/* Now check for time overlap.				*/
		hr_u_start = atoi (hr.unix_start_time);
		hr_u_end = atoi (hr.unix_end_time);
		if (WINDOW_INTERSECT(u_start, u_end, hr_u_start, hr_u_end)) {
		    /*  Have a window to process.			*/
		    if ((pt = parse_date (hr.start_time)) == NULL) {
			error("parsing start_time"); exit(1); }
		    hr_start_time = *pt;
		    if ((pt = parse_date (hr.end_time)) == NULL) {
			error("end_time"); exit(1); }
		    hr_end_time = *pt;

		    /* Construct the request record from the holdings	*/
		    /* record and the the request time.			*/
		    req_start_time =  (tdiff (start_time,hr_start_time) >= 0) 
		      ? start_time : hr_start_time;
		    req_end_time =  (tdiff (end_time,hr_end_time) < 0) 
		      ? end_time : hr_end_time;
		    hr_line = gen_req (hr, req_start_time, req_end_time);
		    puts(hr_line);
		}
		if (matched) break;
	    }
	    if (matched) break;
	}
    }
    return(0);
}

/************************************************************************/
/*  gen_req:								*/
/*	Generate request from holdings record and request file.		*/
/************************************************************************/
char * gen_req (HOLDING_REC hr, INT_TIME req_start_time, INT_TIME req_end_time)
{
    char asc_req_start_time[40], asc_req_end_time[40];
    SDR_TIME sdr_time;
    static char hr_line[256];
    char *location;

    /* Convert times to ascii. */
    strcpy(asc_req_start_time,time_to_str(req_start_time,JULIANC_FMT_1));
    strcpy(asc_req_end_time,time_to_str(req_end_time,JULIANC_FMT_1));
    location = (strcasecmp(hr.dcc,"BK") == 0) ? "  " : hr.dcc;

    sprintf (hr_line,"%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%",
	     hr.station, hr.network, hr.channel, location, 
	     hr.start_time, hr.end_time,
	     trim(hr.data_filename), trim(hr.header_dir), hr.operator,
	     asc_req_start_time, asc_req_end_time);
    return (hr_line);
}
